﻿' RotonVideoGDI
'
' A video plugin that emulates a standard textmode display which outputs to a standard GDI bitmap.

Imports System.Drawing
Imports System.Drawing.Imaging
Imports System.Runtime.InteropServices


Public Class VideoGDI
    Implements Video

    Protected Shared ReadOnly Encoding As System.Text.Encoding = System.Text.Encoding.GetEncoding(437)

    Protected Width As Integer
    Protected Height As Integer
    Protected ViewportWidth As Integer
    Protected ViewportHeight As Integer
    Protected CameraX As Integer
    Protected CameraY As Integer
    Protected FontHeight As Integer
    Protected FontWidth As Integer
    Protected Font(,,) As Int32
    Protected FontI(,,) As Int32
    Protected _Palette() As Int32
    Protected Grid(,) As AnsiChar
    Protected Overlay(,) As AnsiChar
    Protected Updated(,) As Boolean
    Protected OverlayUpdated(,) As Boolean
    Protected BitObj As Bitmap
    Protected BitData() As Int32
    Protected BitHandle As GCHandle
    Protected WidthMax As Integer
    Protected HeightMax As Integer
    Protected ViewWidthMax As Integer
    Protected ViewHeightMax As Integer
    Protected Stride As Integer
    Protected Blink As Boolean
    Protected Bright As Boolean

    Public Sub New()
        SetPalette({
            Color.FromArgb(255, 0, 0, 0),
            Color.FromArgb(255, 0, 0, &HAA),
            Color.FromArgb(255, 0, &HAA, 0),
            Color.FromArgb(255, 0, &HAA, &HAA),
            Color.FromArgb(255, &HAA, 0, 0),
            Color.FromArgb(255, &HAA, 0, &HAA),
            Color.FromArgb(255, &HAA, &H55, 0),
            Color.FromArgb(255, &HAA, &HAA, &HAA),
            Color.FromArgb(255, &H55, &H55, &H55),
            Color.FromArgb(255, &H55, &H55, &HFF),
            Color.FromArgb(255, &H55, &HFF, &H55),
            Color.FromArgb(255, &H55, &HFF, &HFF),
            Color.FromArgb(255, &HFF, &H55, &H55),
            Color.FromArgb(255, &HFF, &H55, &HFF),
            Color.FromArgb(255, &HFF, &HFF, &H55),
            Color.FromArgb(255, &HFF, &HFF, &HFF)
        })
    End Sub

    Public Overridable Function Bitmap() As Bitmap
        If (BitObj Is Nothing) Then
            CreateBitmap()
        End If
        Return BitObj
    End Function

    Public Overridable Function CharAt(ByVal X As Integer, ByVal Y As Integer) As AnsiChar Implements Video.CharAt
        Return Grid(X, Y)
    End Function

    Public Overridable Sub ClearOverlay() Implements Video.ClearOverlay
        Dim X As Integer
        Dim Y As Integer
        For X = 0 To ViewWidthMax
            For Y = 0 To ViewHeightMax
                If (Overlay(X, Y).Char <> 0) Then
                    Overlay(X, Y).Char = 0
                    Updated(X + CameraX, Y + CameraY) = False
                End If
            Next
        Next
    End Sub

    Protected Overridable Sub CreateBitmap()
        Dim BitmapWidth As Integer
        Dim BitmapHeight As Integer

        BitmapWidth = ViewportWidth * FontWidth
        BitmapHeight = ViewportHeight * FontHeight
        If (Not BitHandle.IsAllocated) Then
            ReDim BitData(((ViewportWidth * FontWidth) * (ViewportHeight * FontHeight)) - 1)
            BitHandle = GCHandle.Alloc(BitData, GCHandleType.Pinned)
            BitObj = New Bitmap(ViewportWidth * FontWidth, ViewportHeight * FontHeight, 4 * ViewportWidth * FontWidth, PixelFormat.Format32bppPArgb, BitHandle.AddrOfPinnedObject)
        End If
    End Sub

    Protected Overridable Sub DisposeBitmap()
        If (BitHandle.IsAllocated) Then
            BitObj = Nothing
            BitHandle.Free()
            BitHandle = Nothing
            BitData = Nothing
        End If
    End Sub

    Public Overridable Function ExportTile(ByVal AC As AnsiChar) As Bitmap
        Dim Ch As Integer = AC.Char
        Dim Cl As Integer = AC.Color
        Dim Offset As Integer = 0
        Dim ForeColor As Integer
        Dim BackColor As Integer
        Dim TileBitmap As New Bitmap(FontWidth, FontHeight, PixelFormat.Format32bppPArgb)
        Dim TileBits((FontWidth * FontHeight) - 1) As Int32
        Dim TileBMD As BitmapData = TileBitmap.LockBits(New Rectangle(0, 0, FontWidth, FontHeight), ImageLockMode.WriteOnly, PixelFormat.Format32bppPArgb)

        ForeColor = _Palette(Cl And &HF)
        BackColor = _Palette((Cl And &HF0) >> 4)
        For YIndex = 0 To HeightMax
            For XIndex = 0 To WidthMax
                TileBits(Offset + XIndex) = (Font(Ch, XIndex, YIndex) And ForeColor) Or (FontI(Ch, XIndex, YIndex) And BackColor)
            Next
            Offset += FontWidth
        Next

        Marshal.Copy(TileBits, 0, TileBMD.Scan0, TileBits.Length)
        TileBitmap.UnlockBits(TileBMD)
        Return TileBitmap
    End Function

    Public Overridable ReadOnly Property Palette() As Color()
        Get
            Dim Index As Integer
            Dim Ret As New List(Of Color)

            For Index = 0 To 15
                Ret.Add(Color.FromArgb(_Palette(Index)))
            Next

            Return Ret.ToArray
        End Get
    End Property

    Public Overridable Sub Plot(ByVal X As Integer, ByVal Y As Integer, ByVal AC As AnsiChar) Implements Video.Plot
        If (X >= 0) AndAlso (X < Width) AndAlso (Y >= 0) AndAlso (Y < Height) Then
            Grid(X, Y) = AC
            Updated(X, Y) = False
        End If
    End Sub

    Public Overridable Sub PlotOverlay(ByVal X As Integer, ByVal Y As Integer, ByVal AC As AnsiChar) Implements Video.PlotOverlay
        If (X >= 0) AndAlso (X < ViewportWidth) AndAlso (Y >= 0) AndAlso (Y < ViewportHeight) Then
            Overlay(X, Y) = AC
            OverlayUpdated(X, Y) = False
            Updated(X + CameraX, Y + CameraY) = False
        End If
    End Sub

    Public Sub SetBrightMode(BM As Boolean) Implements Video.SetBrightMode
        Bright = BM
        UpdateBlinks()
    End Sub

    Public Overridable Sub SetCamera(ByVal X As Integer, ByVal Y As Integer)
        Dim XIndex As Integer
        Dim YIndex As Integer
        Dim XEnd As Integer
        Dim YEnd As Integer
        Dim LastCameraX As Integer
        Dim LastCameraY As Integer

        LastCameraX = CameraX
        LastCameraY = CameraY
        CameraX = X
        CameraY = Y
        If (CameraX < 0) Then
            CameraX = 0
        ElseIf (CameraX > (Width - ViewportWidth)) Then
            CameraX = (Width - ViewportWidth)
        End If
        If (CameraY < 0) Then
            CameraY = 0
        ElseIf (CameraY > (Height - ViewportHeight)) Then
            CameraY = (Height - ViewportHeight)
        End If
        If (CameraX <> LastCameraX) OrElse (CameraY <> LastCameraY) Then
            XEnd = CameraX + (ViewportWidth - 1)
            YEnd = CameraY + (ViewportHeight - 1)
            For XIndex = CameraX To XEnd
                For YIndex = CameraY To YEnd
                    Updated(XIndex, YIndex) = False
                Next
            Next
        End If
    End Sub

    Public Overridable Sub SetFont(ByVal F() As Byte)
        Dim X As Integer
        Dim Y As Integer
        Dim C As Integer
        Dim V As Byte
        Dim Offset As Integer

        FontWidth = 8
        FontHeight = F.Length \ 256
        WidthMax = FontWidth - 1
        HeightMax = FontHeight - 1

        ReDim Font(255, FontWidth - 1, FontHeight - 1)
        ReDim FontI(255, FontWidth - 1, FontHeight - 1)

        For C = 0 To 255
            For Y = 0 To FontHeight - 1
                V = F(Offset)
                Offset += 1
                For X = 0 To FontWidth - 1
                    If ((V And 128) = 0) Then
                        Font(C, X, Y) = &HFF000000
                        FontI(C, X, Y) = &HFFFFFFFF
                    Else
                        Font(C, X, Y) = &HFFFFFFFF
                        FontI(C, X, Y) = &HFF000000
                    End If
                    V <<= 1
                Next
            Next
        Next

        DisposeBitmap()
    End Sub

    Public Overridable Sub SetPalette(ByVal P() As Color)
        ReDim _Palette(P.Length - 1)
        Dim Index As Integer

        For Index = 0 To P.Length - 1
            _Palette(Index) = P(Index).ToArgb
        Next
    End Sub

    Public Overridable Sub SetSize(ByVal W As Integer, ByVal H As Integer) Implements Video.SetSize
        ReDim Grid(W - 1, H - 1)
        ReDim Updated(W - 1, H - 1)
        Width = W
        Height = H
        DisposeBitmap()
    End Sub

    Public Overridable Sub SetViewport(ByVal W As Integer, ByVal H As Integer) Implements Video.SetViewport
        ViewportWidth = W
        ViewportHeight = H
        ViewWidthMax = ViewportWidth - 1
        ViewHeightMax = ViewportHeight - 1
        ReDim Overlay(W, H)
        ReDim OverlayUpdated(W, H)
    End Sub

    Public Overridable Sub ToggleBlink() Implements Video.ToggleBlink
        Blink = Not Blink
        If (Not Bright) Then
            UpdateBlinks()
        End If
    End Sub

    Private Sub UpdateBlinks()
        Dim Column As Integer
        Dim Row As Integer

        For Row = 0 To (Height - 1)
            For Column = 0 To (Width - 1)
                If ((Grid(Column, Row).Color And &H80) = &H80) Then
                    Updated(Column, Row) = False
                End If
            Next
        Next
    End Sub

    Public Overridable Sub Update()
        Dim Success As Boolean
        Dim AC As AnsiChar
        Dim X As Integer
        Dim Y As Integer
        Dim OverlayX As Integer
        Dim OverlayY As Integer
        Dim Offset As Integer
        Dim XIndex As Integer
        Dim YIndex As Integer
        Dim Ch As Integer
        Dim ForeColor As Integer
        Dim BackColor As Integer
        Dim ViewStartX As Integer = CameraX
        Dim ViewEndX As Integer = CameraX + ViewWidthMax
        Dim ViewStartY As Integer = CameraY
        Dim ViewEndY As Integer = CameraY + ViewHeightMax
        Stride = FontWidth * ViewportWidth

        If (BitObj Is Nothing) Then
            CreateBitmap()
            If (BitObj Is Nothing) Then
                Exit Sub
            End If
        End If

        For X = ViewStartX To ViewEndX
            For Y = ViewStartY To ViewEndY
                OverlayX = X - ViewStartX
                OverlayY = Y - ViewStartY
                If (Overlay(OverlayX, OverlayY).Char <> 0) Then
                    If (Not OverlayUpdated(OverlayX, OverlayY)) Then
                        AC = Overlay(OverlayX, OverlayY)
                        Success = True
                        OverlayUpdated(OverlayX, OverlayY) = True
                    Else
                        Success = False
                    End If
                ElseIf (Not Updated(X, Y)) Then
                    AC = Grid(X, Y)
                    Success = True
                    Updated(X, Y) = True
                Else
                    Success = False
                End If
                If (Success) Then
                    With AC
                        Ch = .Char
                        If (Bright) Then
                            BackColor = _Palette((.Color And &HF0) >> 4)
                            ForeColor = _Palette(.Color And &HF)
                        Else
                            BackColor = _Palette((.Color And &H70) >> 4)
                            If (Not Blink) AndAlso ((.Color And &H80) = &H80) Then
                                ForeColor = BackColor
                            Else
                                ForeColor = _Palette(.Color And &HF)
                            End If
                        End If
                        Offset = ((X - CameraX) * FontWidth) + (((Y - CameraY) * FontHeight) * Stride)
                        For YIndex = 0 To HeightMax
                            For XIndex = 0 To WidthMax
                                BitData(Offset + XIndex) = (Font(Ch, XIndex, YIndex) And ForeColor) Or (FontI(Ch, XIndex, YIndex) And BackColor)
                            Next
                            Offset += Stride
                        Next
                    End With
                End If
            Next
        Next
    End Sub

    Public Overridable ReadOnly Property ViewWidth As Integer
        Get
            Return ViewportWidth
        End Get
    End Property

    Public Overridable ReadOnly Property ViewHeight As Integer
        Get
            Return ViewportHeight
        End Get
    End Property

    Public Overridable Sub Write(ByVal X As Integer, ByVal Y As Integer, ByVal S As String, ByVal Col As Integer) Implements Video.Write
        Dim Chars() As Byte
        Dim Index As Integer

        Chars = Encoding.GetBytes(S)
        For Index = 0 To Chars.Length - 1
            Plot(X, Y, New AnsiChar(Chars(Index), Col))
            X += 1
        Next
    End Sub

    Public Overridable Sub WriteOverlay(ByVal X As Integer, ByVal Y As Integer, ByVal S As String, ByVal Col As Integer) Implements Video.WriteOverlay
        Dim Chars() As Byte
        Dim Index As Integer

        Chars = Encoding.GetBytes(S)
        For Index = 0 To Chars.Length - 1
            PlotOverlay(X, Y, New AnsiChar(Chars(Index), Col))
            X += 1
        Next
    End Sub

    Protected Overrides Sub Finalize()
        DisposeBitmap()
        MyBase.Finalize()
    End Sub

    Public Sub CancelUpdate(X As Integer, Y As Integer) Implements Video.CancelUpdate
        Updated(X, Y) = True
    End Sub
End Class
